/*
Subject: general purpose telnet server.
This is a telnet bouncer that some people have expressed interest in.
It's uses are utility and privacy related. It should compile easily on
most OS's striaght out of it's wrapper as such.

Mark
mark@cairo.anu.edu.au
*/

/* modified by Roger Espel LLima 12/3/94: open a new socket instead of
   reusing the same one, so the standard telnet client may be used
   transparently.
   modified by Roger Espel Llima 13/3/94: handle ftp PORT commands 
   easy to add new services, "who" service, and wrote the clients that
   connect to it, get a port and then launch telnet/ftp.

   lots of bug fixes and added tests
   possibility to specify the server's port
   added an error log file
   swapped child/parent after some forks to avoid zombies/double forks
   protocol remark: data may not be sent in advance, i.e a client cannot
   send the host name before the server asks for it; it will FAIL.
   not tried on anything else than SunOS 4.1 and AIX
*/

#include "relay.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>

#ifdef _AIX
#include <sys/select.h>
#endif

#include <fcntl.h>
#include <errno.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <sys/wait.h>

#define    QLEN           5
#define    FTP_PORT 21

char sbuf[16384], cbuf[16384];
struct hostent *myself;
char mynumber[100];
extern int errno;
extern char *sys_errlist[];
int debug=0;

void sigchld() {
  signal(SIGCHLD, sigchld);
  while(waitpid(0, (int *)0, WNOHANG|WUNTRACED)>=0);
}

void doerror(char *message,int nl) {
    if (debug) {
	int fd;
	if ((fd=open(LOGFILE, O_WRONLY|O_CREAT|O_APPEND, 0600))<0) return;
	write(fd,message,strlen(message));
	if (nl>0) write(fd,sys_errlist[nl],strlen(sys_errlist[nl]));
	write(fd,"\n",1);
	close(fd);
    }
}

void wwrite(int fd,char *string) {
    int n,l;
    n=write(fd,string,l=strlen(string));
    if (n!=l) exit(-1);
}

void rread(int fd,char *string,int len) {
    /* reads up to a \n or len chars */
    int n;
    n=read(fd,string,len);
    if (n<1) {
	doerror("read error: connection closed?",0);
   	exit(-1);
    }
}

void check(char *string) {
    char *t;
    t=string;
    while (*t) {
	if (((unsigned int)(*t)<33) || ((unsigned int)(*t)>126)) {
	    doerror("special chars found!",0);
	    exit(-1);
	}
	++t;
    }
}

void clean(char *string) {
    char *t;
    t=string+(strlen(string)-1);
    while (t>=string && (*t=='\r' || *t=='\n' || *t<=' ')) --t;
    *++t='\0';
}

void communicate(int sfd,int cfd,int ftp) {
    char *chead, *ctail, *shead, *stail;
    int num, num2, nfd, spos, cpos;
    int i1,i2,i3,i4,i5,i6,n;
    char t;

    extern int errno;
    fd_set rd, wr;

    int ftp_srv_fd, ftp_rem_fd, ftp_rem_len;
    struct sockaddr_in ftp_rem_addr, ftp_srv_addr;
    int length;
    int ftpdest;

    struct sockaddr_in sa;
    char string[100];

    struct itimerval itime;

    /* sfd = destination ; cfd = client */

    chead = ctail = cbuf;
    cpos = 0;
    shead = stail = sbuf;
    spos = 0;
    while (1) {
        FD_ZERO(&rd);
        FD_ZERO(&wr);
        if (spos < sizeof(sbuf)-1) FD_SET(sfd, &rd);
        if (ctail > chead) FD_SET(sfd, &wr);
        if (cpos < sizeof(cbuf)-1) FD_SET(cfd, &rd);
        if (stail > shead) FD_SET(cfd, &wr);
        nfd = select(256, &rd, &wr, 0, 0);
        if (nfd <= 0) continue;
        if (FD_ISSET(sfd, &rd)) {
            num=read(sfd,stail,sizeof(sbuf)-spos);
            if ((num==-1) && (errno != EWOULDBLOCK)) return;
            if (num==0) {
#ifdef FNDELAY
		fcntl(sfd,F_SETFL,fcntl(sfd,F_GETFL,0)&~FNDELAY);
		fcntl(cfd,F_SETFL,fcntl(cfd,F_GETFL,0)&~FNDELAY);
#endif
		if (ctail!=chead) write(sfd,chead,ctail-chead);
		if (stail!=shead) write(cfd,shead,stail-shead);
		write(cfd,string,0);
		return;
	    }
	    
            if (num>0) {
                spos += num;
                stail += num;
                if (!--nfd) continue;
            }
        }
        if (FD_ISSET(cfd, &rd)) {
            num=read(cfd,ctail,sizeof(cbuf)-cpos);
            if ((num==-1) && (errno != EWOULDBLOCK)) return;
            if (num==0) {
#ifdef FNDELAY
		fcntl(sfd,F_SETFL,fcntl(sfd,F_GETFL,0)&~FNDELAY);
		fcntl(cfd,F_SETFL,fcntl(cfd,F_GETFL,0)&~FNDELAY);
#endif
		if (ctail!=chead) write(sfd,chead,ctail-chead);
		if (stail!=shead) write(cfd,shead,stail-shead);
		write(sfd,string,0);
		return;
	    }
		
            if (num>0) {
		/* test here if ftp!=0 && it is a PORT command */
		if (ftp) n=sscanf(ctail," PORT %i ,%i ,%i ,%i ,%i ,%i%c",&i1,&i2,&i3,&i4,&i5,&i6,&t); else n=0;
		if (n==7 && (t==(char)13 || t==(char)10)) {
		    if (ctail!=chead) { /* write pending data to the server */
			num2=write(sfd,chead,ctail-chead);
			if ((num2==-1) && (errno != EWOULDBLOCK)) return;
			chead+=num2;
		    }
		    if (ctail==chead) { /* if we managed to write it */
			while (*ctail!=t) {
			    ctail++;
			    num-- ;
			}
			chead = ctail = cbuf;
			cpos = 0;
			num = 0;
			/* skip the PORT command (and whatever came after it)
			   (bug?) */
			
			/* open & handle PORT connection */
			sprintf(string,"%i.%i.%i.%i",i1,i2,i3,i4);
			sa.sin_addr.s_addr = inet_addr(string);
			sa.sin_family = AF_INET;
			sa.sin_port = htons((unsigned) (i5<<8)+i6);

			memset((char *)&ftp_rem_addr, 0, sizeof(ftp_rem_addr));
			memset((char *)&ftp_srv_addr, 0, sizeof(ftp_srv_addr));
			ftp_srv_addr.sin_family = AF_INET;
			ftp_srv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
			ftp_srv_addr.sin_port = 0;
			ftp_srv_fd = socket(PF_INET, SOCK_STREAM, 0);
			if (bind(ftp_srv_fd, (struct sockaddr *) &ftp_srv_addr, sizeof(ftp_srv_addr)) == -1) {
			    doerror("bind: ",errno);
			    exit(-1);
			}
			length=sizeof(ftp_srv_addr);
			if (getsockname(ftp_srv_fd, (struct sockaddr *) &ftp_srv_addr, &length) == -1 ) {
			    doerror("getsockname: ",errno);
			    exit(-1);
			}

			sprintf(string,"PORT %s,%i,%i\r\n",mynumber,(int)ntohs(ftp_srv_addr.sin_port)>>8,(int)ntohs(ftp_srv_addr.sin_port) & 0xff);
			num2=write(sfd,string,strlen(string));
			if (num2!=strlen(string))  {
			    doerror("ftp: write: ",errno);
			    exit(-1);
			}

			switch (fork()) {
			  case -1:
			    doerror("ftp: fork: ",errno);
			    exit(-1);
			  case 0:
    /* make sure the PARENT takes the new ftp connection and the CHILD
       goes on with the ftp control socket, so we don't have
       a zombie for the rest of the ftp session */
			    close(ftp_srv_fd);
			    itime.it_interval.tv_sec=0;
			    itime.it_interval.tv_usec=0;
			    itime.it_value.tv_sec=21600;
   /* restore the connection time limit: 6 hours in case the client hangs */
			    itime.it_value.tv_usec=0;
			    setitimer(ITIMER_REAL,&itime,NULL);
			    break;

			  default:

			    listen(ftp_srv_fd, 1);
			    ftp_rem_len=sizeof(ftp_rem_addr);

			    itime.it_interval.tv_sec=0;
			    itime.it_interval.tv_usec=0;
			    itime.it_value.tv_sec=150;
			    itime.it_value.tv_usec=0;
			    setitimer(ITIMER_REAL,&itime,NULL);
			
			    ftp_rem_fd=accept(ftp_srv_fd, (struct sockaddr *) &ftp_rem_addr, &ftp_rem_len);
			    if (ftp_rem_fd < 0) {
				doerror("connect: accept: ",errno);
				exit(-1);
			    }

			    itime.it_interval.tv_sec=0;
			    itime.it_interval.tv_usec=0;
			    itime.it_value.tv_sec=21600;
			    itime.it_value.tv_usec=0;
			    setitimer(ITIMER_REAL,&itime,NULL);
			    
			    if ((ftpdest = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
				doerror("ftp: socket: ",errno);
				close(sfd);
				close(cfd);
				exit(1);
			    }
			    if (connect(ftpdest, (struct sockaddr *) &sa, sizeof(sa))<0) {
				doerror("ftp: connect: ",errno);
				close(sfd);
				close(cfd);
				exit(1);
			    }
#ifdef FNDELAY
			    fcntl(ftp_rem_fd,F_SETFL,fcntl(ftp_rem_fd,F_GETFL,0)|FNDELAY);
			    fcntl(ftpdest,F_SETFL,fcntl(ftpdest,F_GETFL,0)|FNDELAY);
#else
			    fcntl(ftp_rem_fd,F_SETFL,O_NDELAY);
			    fcntl(ftpdest,F_SETFL,O_NDELAY);
#endif
			    close(ftp_srv_fd);
			    close(sfd);
			    close(cfd);
			    communicate(ftpdest,ftp_rem_fd,0);
			    close(ftpdest);
			    close(ftp_rem_fd);
			    exit(0);
			}
		    }
		}
                cpos+=num;
                ctail+=num;
                if (!--nfd) continue;
            }
        }
        if (FD_ISSET(sfd, &wr)) {
            num=write(sfd,chead,ctail-chead);
            if ((num==-1) && (errno != EWOULDBLOCK)) return;
            if (num>0) {
                chead += num;
                if (chead == ctail) {
                    chead = ctail = cbuf;
                    cpos = 0;
                }
                if (!--nfd) continue;
            }
        }
        if (FD_ISSET(cfd, &wr)) {
            num=write(cfd,shead,stail-shead);
            if ((num==-1) && (errno != EWOULDBLOCK)) return;
            if (num>0) {
                shead += num;
                if (shead == stail) {
                    shead = stail = sbuf;
                    spos = 0;
                }
                if (!--nfd) continue;
            }
        }
    }
}

void telcli(int source) {
    int new_srv_fd,new_rem_fd,new_rem_len;
    struct sockaddr_in new_rem_addr,new_srv_addr;

    int dest;
    int length;
    int ftp;

    struct sockaddr_in sa;
    struct hostent *hp;
    char gethost[100];
    char getport[100];
    char string[100];
    char command[200];
    int servfound=0;
    int n;

    struct itimerval itime;

    FILE *finger;
    unsigned char *t;

    itime.it_interval.tv_sec=0;
    itime.it_interval.tv_usec=0;
    itime.it_value.tv_sec=21600;
    /* arbitrary connection time limit: 6 hours in case the client hangs */
    itime.it_value.tv_usec=0;
    setitimer(ITIMER_REAL,&itime,NULL);

    /*
  ---- add any authentification here -----
    */
    memset(string, 0, 100);
    rread(source,string,95);
    clean(string);
    if (strcmp(string,IDENTIFICATION)) {
      close(source);
      exit(1);
    }

    wwrite(source,"Ok\r\n");
    do {
	wwrite(source,"Service? ");
	memset(string, 0, 100);
	rread(source,string,10);
	clean(string);
	if (strlen(string)==0) continue;
	if (strcasecmp(string,"help")==0) {
	    wwrite(source,"Available services: help, finger, telnet, ftp, quit\r\n");
	    continue;
	}
	else if (strcasecmp(string,"finger")==0) {
	    servfound=1;
	    break;
	}
	else if (strcasecmp(string,"telnet")==0) {
	    servfound=2;
	    ftp=0;
	    break;
	}
	else if (strcasecmp(string,"ftp")==0) {
	    servfound=2;
	    ftp=1;
	    break;
	}
	else if (strcasecmp(string,"quit")==0) {
	    close(source);
	    exit(0);
	}
	else {
	    sprintf(getport,"Unknown service '%s' - try again.\r\n",string);
	    wwrite(source,getport);
	}
    } while (servfound==0);

    if (servfound==1) {
        memset(gethost, 0, 100);
        wwrite(source,"Who: ");
        rread(source,gethost,90);
        clean(gethost);
        t=gethost;
        while(*t) {
            if ((*t<'a' || *t>'z') && (*t<'A' || *t>'Z') &&
                (*t<'0' || *t>'9') && *t!='.' && *t!='-' &&
                *t!='@') {
                t=NULL;
                break;
            }
            ++t;
        }
        if (t==NULL) {
            wwrite(source,"** bad user name\r\n");
            close(source);
            exit(-1);
        }
        sprintf(command,"%s %s",FINGER,gethost);
        itime.it_interval.tv_sec=0;
        itime.it_interval.tv_usec=0;
        itime.it_value.tv_sec=120;
        /* arbitrary connection time limit: 2 minutes */
        itime.it_value.tv_usec=0;
        setitimer(ITIMER_REAL,&itime,NULL);

        if ((finger=popen(command,"r"))==NULL) {
            wwrite(source,"** can't fork\r\n");
            close(source);
            exit(-1);
        }
        do {
            n=fread(sbuf,1,16000,finger);
            if (n>0) n=write(source,sbuf,n);
        } while (n>0);
        pclose(finger);
        close(source);
        exit(0);
    }

    memset(gethost, 0, 100);
    wwrite(source,"Host: ");
    rread(source,gethost,90);
    clean(gethost);
    if (strlen(gethost)<1) {
	wwrite(source,"** invalid host\r\n");
	close(source);
	exit(-1);
    }
    hp=gethostbyname(gethost);
    if (hp) {
      sa.sin_addr=*(struct in_addr *)(hp->h_addr_list[0]);
    } else {
      if (inet_addr(gethost) == -1) {
	  wwrite(source,"** address not found\r\n");
	  close(source);
	  exit(-1);
      } else {
	  sa.sin_addr.s_addr = inet_addr(gethost);
      }
    }
    sa.sin_family=AF_INET;
    memset(getport, 0, 100);
    wwrite(source,"Port: ");
    rread(source,getport,50);
    clean(getport);
    if (strlen(getport)<1) {
	wwrite(source,"** invalid port\r\n");
	close(source);
	exit(-1);
    }
    if ((strcasecmp(getport,"ftp"))==0)
	sa.sin_port=htons(FTP_PORT);
    else
	sa.sin_port=htons((unsigned) atoi(getport));

    if (sa.sin_port==0) {
	wwrite(source,"** invalid port number\r\n");
	close(source);
	exit(-1);
    }
/*  sprintf(string, "Trying %s...\r\n", (char *) inet_ntoa(sa.sin_addr));
    wwrite(source,string);  */

    memset((char *)&new_rem_addr, 0, sizeof(new_rem_addr));
    memset((char *)&new_srv_addr, 0, sizeof(new_srv_addr));
    new_srv_addr.sin_family = AF_INET;
    new_srv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    new_srv_addr.sin_port = 0;
    new_srv_fd = socket(PF_INET, SOCK_STREAM, 0);
    if (bind(new_srv_fd, (struct sockaddr *) &new_srv_addr, sizeof(new_srv_addr)) == -1) {
        doerror("bind: ",errno);
        exit(-1);
    }
    length=sizeof(new_srv_addr);
    if (getsockname(new_srv_fd, (struct sockaddr *) &new_srv_addr, &length) == -1 ) {
	doerror("getsockname: ",errno);
	exit(-1);
    }

    sprintf(string,"Connect to port %d.\r\n",ntohs(new_srv_addr.sin_port));
    wwrite(source,string);
    close(source);

    listen(new_srv_fd, 1);
    new_rem_len = sizeof(new_rem_addr);

    itime.it_interval.tv_sec=0;
    itime.it_interval.tv_usec=0;
    itime.it_value.tv_sec=60;
    itime.it_value.tv_usec=0;
    setitimer(ITIMER_REAL,&itime,NULL);

    new_rem_fd=accept(new_srv_fd, (struct sockaddr *) &new_rem_addr, &new_rem_len);
    if (new_rem_fd < 0) exit(-1);

    itime.it_interval.tv_sec=0;
    itime.it_interval.tv_usec=0;
    itime.it_value.tv_sec=21600;
    /* arbitrary connection time limit: 6 hours (in case the client hangs) */
    itime.it_value.tv_usec=0;
    setitimer(ITIMER_REAL,&itime,NULL);

    if ((dest=socket(AF_INET,SOCK_STREAM,0)) < 0) {
	close(new_rem_fd);
	close(new_srv_fd);
        exit(1);
    }
    if (connect(dest,(struct sockaddr *)&sa,sizeof(sa))<0) {
	close(new_rem_fd);
	close(new_srv_fd);
	close(dest);
	exit(-1);
    }

#ifdef FNDELAY
    fcntl(new_rem_fd,F_SETFL,fcntl(new_rem_fd,F_GETFL,0)|FNDELAY);
    fcntl(dest,F_SETFL,fcntl(dest,F_GETFL,0)|FNDELAY);
#else
    fcntl(new_rem_fd,F_SETFL,O_NDELAY);
    fcntl(dest,F_SETFL,O_NDELAY);
#endif
    close(new_srv_fd);
    communicate(dest,new_rem_fd,ftp);
    close(dest);
    close(new_rem_fd);
    exit(0);
}

int main(int argc,char *argv[]) {
    int srv_fd,rem_fd,rem_len;
    int myport=SERV_TCP_PORT;
    char name[100],*temp;
    struct sockaddr_in rem_addr,srv_addr;

    /* only command-line errors go to stderr */
    while (argc>1) {
	if (strcmp(argv[1],"-d")==0) {
	    debug=1;
	    argv++;
	    argc--;
	    continue;
	}
	if (strcmp(argv[1],"-p")==0 && argc>2) {
	    if ((myport=atoi(argv[2]))==0) {
		fprintf(stderr,"Bad port number.\n");
		exit(-1);
	    }
	    argv+=2;
	    argc-=2;
	    continue;
	}
	fprintf(stderr,"Use: %s [ -d ] [ -p port ]\n",argv[0]);
	exit(-1);
    }

    if (gethostname(name,100)<0) {
	doerror("gethostname: ",errno);
	exit(1);
    }
    myself=gethostbyname(name);
    if (myself)
	strcpy(mynumber, inet_ntoa(*(struct in_addr *)(myself->h_addr_list[0])));
    else {
	doerror("gethostname: ",errno);
	exit(1);
    }
    temp=mynumber;
    while (*temp)  {
	if (*temp=='.') *temp=',';
	++temp;
    }

    memset((char *) &rem_addr, 0, sizeof(rem_addr));
    memset((char *) &srv_addr, 0, sizeof(srv_addr));
    srv_addr.sin_family=AF_INET;
    srv_addr.sin_addr.s_addr=htonl(INADDR_ANY);
    srv_addr.sin_port=htons(myport);
    srv_fd=socket(PF_INET,SOCK_STREAM,0);
    if (bind(srv_fd,&srv_addr,sizeof(srv_addr))==-1) {
	if (!debug) perror("bind: ");
        exit(-1);
    }
    listen(srv_fd,QLEN);
    close(0); close(1); close(2);
#ifdef TIOCNOTTY
    if ((rem_fd=open("/dev/tty", O_RDWR)) >= 0) {
        ioctl(rem_fd, TIOCNOTTY, (char *)0);
        close(rem_fd);
    }
#endif
    if (fork()) exit(0);
    signal(SIGCHLD, sigchld);
    while (1) {
	rem_len=sizeof(rem_addr);
	rem_fd=accept(srv_fd,&rem_addr,&rem_len);
	if (rem_fd<0) {
	  if (errno==EINTR) continue;
	  exit(-1);
        }
	switch(fork()) {
	  case -1:
	    doerror("fork: ",errno);
	    break;
	  case 0:                           /* child process */
	    close(rem_fd);
	 /* if (fork()) exit(0);   */       /* let init worry about children */
	    break;
	  default:			    /* parent process */
#ifdef SALLE_S
	    strcpy(name,inet_ntoa(rem_addr.sin_addr));
	    n=sscanf(name,"129.199.129.%i",&clientmachine);
	    if ((!n) && strcmp(name,"129.199.1.22")) {
		sprintf(string,"Connection not allowed from %s\r\n",name);
		wwrite(rem_fd,string);
		close(rem_fd);
		close(srv_fd);
		exit(-1);
	    }
#endif
	    close(srv_fd);                  /* close original socket */
	    telcli(rem_fd);                 /* process the request */
	    close(rem_fd);
	    exit(0);
	}
    }
}

