/* relay finger
   by Roger Espel Llima (roger.espel.llima@ens.fr) '94
*/

#include "relay.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>

void clean(char *string) {
    char *t;
    t=string+(strlen(string)-1);
    while (t>=string && (*t=='\r' || *t=='\n' || *t==' ')) --t;
    *++t='\0';
}

void proterror(int n,char *string) {
    fprintf(stderr,"protocol error %i: '%s'\n",n,string);
    exit(-1);
}

void wwrite(int fd,char *string) {
    int n;
    n=write(fd,string,strlen(string));
    if (n!=strlen(string)) {
	fprintf(stderr,"write error: connection closed?\n");
	exit(-1);
    }
}

void rread(int fd,char *string,int len) {
    int n;
    n=read(fd,string,len);
    if (n<1) {
	fprintf(stderr,"read error: connection closed?\n");
	exit(-1);
    }
}

void check(char *string) {
    char *t;
    t=string;
    while (*t) {
	if (((unsigned int)(*t)<33) || ((unsigned int)(*t)>126)) {
	    fprintf(stderr,"special chars in command line!\n");
	    exit(-1);
	}
	++t;
    }
}

int main(int argc,char **argv) {
    struct sockaddr_in s;
    struct hostent *hp;
    int sock;
    int relayport=SERV_TCP_PORT;
    char string[100],relay[100],buffer[2000],user[100];
    int n;
    char *myname;
    
    myname=argv[0];
    strcpy(relay,RELAY);
    while (argc>2 && (strcmp(argv[1],"-r")==0 || strcmp(argv[1],"-p")==0)) {
	if (strcmp(argv[1],"-r")==0) {
	    if (strlen(argv[2])>70) {
		fprintf(stderr,"Relay machine name too long.\n");
		exit(-1);
	    }
	    check(argv[2]);
	    strcpy(relay,argv[2]);
	} else {
	    if (strlen(argv[2])>7) {
		fprintf(stderr,"Port number too long!\n");
		exit(-1);
	    }	
	    check(argv[2]);
	    relayport=atoi(argv[2]);
	    if (!relayport) {
		fprintf(stderr,"Invalid relay port.\n");
		exit(-1);
	    }
	}
	argv++;
	argv++;
	argc--;
	argc--;
    }
   
    if (argc!=1 && argc!=2) {
	fprintf(stderr,"Usage: %s [ -r relaymachine ] [ -p relayport ] [ user ]\n",myname);
	exit(-1);
    }

    if (argc==2 && strlen(argv[1])>80) {
	fprintf(stderr,"Username too long.\n");
	exit(-1);
    }
    
    hp=gethostbyname(relay);
    if (hp) {
      s.sin_addr=*(struct in_addr *)(hp->h_addr_list[0]);
    } else {
	fprintf(stderr,"Can't access relay machine.\n");
	exit(-1);
    }
    s.sin_family=AF_INET;
    s.sin_port=relayport;

    sock=socket(PF_INET,SOCK_STREAM,0);
    if (connect(sock,&s,sizeof(s))<0) {
	fprintf(stderr,"Can't connect to relay daemon.\n");
	exit(-1);
    }

    wwrite(sock,IDENTIFICATION "\r\n");
    memset(string, 0, 100);
    rread(sock,string,4);
    clean(string);
    if (strncmp(string,"Ok",2)) proterror(0,string);

    memset(string, 0, 100);
    rread(sock,string,50);
    clean(string);
    if (strcasecmp(string,"Service?")) proterror(1,string);
    wwrite(sock,"finger\r\n");
    memset(string, 0, 100);
    rread(sock,string,50);
    clean(string);
    if (strcasecmp(string,"Who:")) proterror(2,string);
    if (argc==1)
	wwrite(sock,"\r\n");
    else {
	sprintf(user,"%s\r\n",argv[1]);
	wwrite(sock,user);
    }
    do	{
	n=read(sock,buffer,sizeof(buffer));
	if (n>0) write(1,buffer,n);
    } while (n>0);
    close(sock);
}

