/* telnet client: calls a relay running telrelayftp
   by Roger Espel Llima (roger.espel.llima.ens.fr) '94
   (executes ftp if called under the name *ftp)
   (prints Ok machinename portnumber if called under the name *open)

   
   BUGS: does not allow passing options to ftp
   LIMITATIONS: if new connections are opened from within ftp/telnet,
                they are not relayed
   not tried on anything else than SunOS 4.1.3
*/


#include "relay.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>

#define    FTP_PORT "21\r\n"
#define	   TELNET_PORT "23\r\n"

void clean(char *string) {
    char *t;
    t=string+(strlen(string)-1);
    while (t>=string && (*t=='\r' || *t=='\n' || *t==' ')) --t;
    *++t='\0';
}

void proterror(int n,char *string) {
    fprintf(stderr,"protocol error %i: '%s'\n",n,string);
    exit(-1);
}

void wwrite(int fd,char *string) {
    int n;
    n=write(fd,string,strlen(string));
    if (n!=strlen(string)) {
	fprintf(stderr,"write error: connection closed?\n");
	exit(-1);
    }
}

void rread(int fd,char *string,int len) {
    int n;
    n=read(fd,string,len);
    if (n<1) {
	fprintf(stderr,"read error: connection closed?\n");
	exit(-1);
    }
}

void check(char *string) {
    char *t;
    t=string;
    while (*t) {
	if (((unsigned int)(*t)<33) || ((unsigned int)(*t)>126)) {
	    fprintf(stderr,"special chars in command line!\n");
	    exit(-1);
	}
	++t;
    }
}

int main(int argc,char **argv) {
    int port;
    struct sockaddr_in s;
    struct hostent *hp;
    int sock;
    int relayport=SERV_TCP_PORT;
    char string[100];
    char relay[100];
    int n,ftp=0,onlyopen=0;
    char *t,*myname;

    strcpy(relay,RELAY);
    myname=*argv;
    while (argc>2 && (strcmp(argv[1],"-r")==0 || strcmp(argv[1],"-p")==0)) {
	if (strcmp(argv[1],"-r")==0) {
	    if (strlen(argv[2])>70) {
		fprintf(stderr,"Relay machine name too long.\n");
		exit(-1);
	    }
	    check(argv[2]);
	    strcpy(relay,argv[2]);
	} else {
	    if (strlen(argv[2])>7) {
		fprintf(stderr,"Port number too long!\n");
		exit(-1);
	    }	
	    check(argv[2]);
	    relayport=atoi(argv[2]);
	    if (!relayport) {
		fprintf(stderr,"Invalid relay port.\n");
		exit(-1);
	    }
	}
	argv++;
	argv++;
	argc--;
	argc--;
    }
   
    if (argc<2 || argc>3) {
	fprintf(stderr,"Usage: %s [ -r relaymachine ] [ -p relayport ] machine [port]\n",myname);
	exit(-1);
    }
    if (strlen(argv[1])>70) {
	fprintf(stderr,"Target machine name too long.\n");
	exit(-1);
    }
    check(argv[1]);
    if (argc==3) {
	if (strlen(argv[2])>7) {
	    fprintf(stderr,"Port number too long.\n");
	    exit(-1);
	}	
	check(argv[2]);
    }

    t=myname;
    if (strlen(t)>2) {
	while (*t) ++t;
	--t;
	if (*(t-2)=='f' && *(t-1)=='t' && *(t)=='p') ++ftp;
    }
    t=myname;
    if (strlen(t)>3) {
	while (*t) ++t;
	--t;
	if (*(t-3)=='o' && *(t-2)=='p' && *(t-1)=='e' && *(t)=='n') ++onlyopen;
    }
    
    hp=gethostbyname(relay);
    if (hp) {
      s.sin_addr=*(struct in_addr *)(hp->h_addr_list[0]);
    } else {
	fprintf(stderr,"Can't access relay machine.\n");
	exit(-1);
    }
    s.sin_family=AF_INET;
    s.sin_port=relayport;

    sock=socket(PF_INET,SOCK_STREAM,0);
    if (connect(sock,&s,sizeof(s))<0) {
	fprintf(stderr,"Can't connect to relay daemon.\n");
	exit(-1);
    }

    wwrite(sock,IDENTIFICATION "\r\n");

    memset(string, 0, 100);
    rread(sock,string,4);
    clean(string);
    if (strncmp(string,"Ok",2)) proterror(0,string);
    
    memset(string, 0, 100);
    rread(sock,string,50);
    clean(string);
    if (strcasecmp(string,"Service?")) proterror(1,string);
    if (ftp) wwrite(sock,"ftp\r\n"); else wwrite(sock,"telnet\r\n");
    memset(string, 0, 100);
    rread(sock,string,50);
    clean(string);
    if (strcasecmp(string,"Host:")) proterror(2,string);
    strcpy(string,argv[1]);
    strcat(string,"\r\n");
    wwrite(sock,string);
    
    memset(string, 0, 100);
    rread(sock,string,50);
    clean(string);
    if (strcasecmp(string,"Port:")) proterror(3,string);
    if (argc==3) {
	strcpy(string,argv[2]);
	strcat(string,"\r\n");
	wwrite(sock,string);
    } else {
	if (ftp) wwrite(sock,FTP_PORT); else wwrite(sock,TELNET_PORT);
    }

    memset(string, 0, 100);
    rread(sock,string,60);
    close(sock);
    clean(string);
    n=sscanf(string,"Connect to port %i.",&port);
    if (n==0) proterror(4,string);
    if (onlyopen) {
	printf("Ok %s %i\n",relay,port);
	exit(0);
    }
    sprintf(string,"%i",port);
    if (ftp) printf("Opening ftp connection.\n");
    else printf("Opening telnet connection.\n");
    execl(ftp?FTP:TELNET,ftp?"ftp":"telnet",relay,string,(char *)0);
    fprintf(stderr,"Can't execute %s\n",ftp?FTP:TELNET);
}

